from typing import List

import httpx
from solana.rpc.async_api import (
    AsyncClient as async_conn,
)  # pylint: disable=unused-import # noqa:F401

from .connection import (
    LIVE_MARKETS_URL,
    TOKEN_MINTS_URL,
    parse_live_markets,
    parse_token_mints,
)
from .market.types import MarketInfo, TokenInfo


async def get_live_markets(httpx_client: httpx.AsyncClient) -> List[MarketInfo]:
    resp = await httpx_client.get(LIVE_MARKETS_URL)
    return parse_live_markets(resp.json())


async def get_token_mints(httpx_client: httpx.AsyncClient) -> List[TokenInfo]:
    resp = await httpx_client.get(TOKEN_MINTS_URL)
    return parse_token_mints(resp.json())



"""Market module to interact with Serum DEX."""
from __future__ import annotations

import itertools
import logging
from typing import List, Union

from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solders.system_program import CreateAccountParams, create_account
from solana.transaction import Transaction
from solders.instruction import Instruction
from solders.rpc.responses import GetMinimumBalanceForRentExemptionResp
from spl.token.constants import ACCOUNT_LEN, TOKEN_PROGRAM_ID, WRAPPED_SOL_MINT
from spl.token.instructions import (
    CloseAccountParams,
    InitializeAccountParams,
    close_account,
    initialize_account,
)

import pyserum.market.types as t
from pyserum import instructions

from ..async_open_orders_account import AsyncOpenOrdersAccount
from ..enums import OrderType, SelfTradeBehavior, Side
from ..open_orders_account import OpenOrdersAccount, make_create_account_instruction
from ._internal.queue import decode_event_queue
from .orderbook import OrderBook
from .state import MarketState

from decimal import Decimal

LAMPORTS_PER_SOL = 1000000000


# pylint: disable=too-many-public-methods
class MarketCore:
    """Represents a Serum Market."""

    logger = logging.getLogger("pyserum.market.Market")

    def __init__(
        self, market_state: MarketState, force_use_request_queue: bool = False
    ) -> None:
        self.state = market_state
        self.force_use_request_queue = force_use_request_queue

    def _use_request_queue(self) -> bool:
        return (
            # DEX Version 1
            self.state.program_id
            == Pubkey.from_string("4ckmDgGdxQoPDLUkDT3vHgSAkzA3QRdNq5ywwY4sUSJn")
            or
            # DEX Version 1
            self.state.program_id
            == Pubkey.from_string("BJ3jrUzddfuSrZHXSCxMUUQsjKEyLmuuyZebkcaFp2fg")
            or
            # DEX Version 2
            self.state.program_id
            == Pubkey.from_string("EUqojwWA2rd19FZrzeBncJsm38Jm1hEhE3zsmX3bRc2o")
            or self.force_use_request_queue
        )

    def support_srm_fee_discounts(self) -> bool:
        raise NotImplementedError("support_srm_fee_discounts not implemented")

    def find_fee_discount_keys(self, owner: Pubkey, cache_duration: int):
        raise NotImplementedError("find_fee_discount_keys not implemented")

    def find_best_fee_discount_key(self, owner: Pubkey, cache_duration: int):
        raise NotImplementedError("find_best_fee_discount_key not implemented")

    def find_quote_token_accounts_for_owner(
        self, owner_address: Pubkey, include_unwrapped_sol: bool = False
    ):
        raise NotImplementedError("find_quote_token_accounts_for_owner not implemented")

    def _parse_bids_or_asks(self, bytes_data: bytes) -> OrderBook:
        return OrderBook.from_bytes(self.state, bytes_data)

    @staticmethod
    def _parse_orders_for_owner(bids, asks, open_orders_accounts) -> List[t.Order]:
        if not open_orders_accounts:
            return []

        all_orders = itertools.chain(bids.orders(), asks.orders())
        open_orders_addresses = {str(o.address) for o in open_orders_accounts}
        orders = [
            o for o in all_orders if str(o.open_order_address) in open_orders_addresses
        ]
        return orders

    def load_base_token_for_owner(self):
        raise NotImplementedError("load_base_token_for_owner not implemented")

    def _parse_fills(self, bytes_data: bytes, limit: int) -> List[t.FilledOrder]:
        events = decode_event_queue(bytes_data, limit)
        return [
            self.parse_fill_event(event)
            for event in events
            if event.event_flags.fill and event.native_quantity_paid > 0
        ]

    def parse_fill_event(self, event: t.Event) -> t.FilledOrder:
        if event.event_flags.bid:
            side = 'bids'

            price_before_fees = (
                event.native_quantity_paid + event.native_fee_or_rebate
                if event.event_flags.maker
                else event.native_quantity_paid - event.native_fee_or_rebate
            )

            price = (price_before_fees * self.state.base_spl_token_multiplier()) / (
                    self.state.quote_spl_token_multiplier() * event.native_quantity_released
            )

            size = event.native_quantity_released / self.state.base_spl_token_multiplier()
        else:
            side = 'asks'

            price_before_fees = (
                event.native_quantity_released - event.native_fee_or_rebate
                if event.event_flags.maker
                else event.native_quantity_released + event.native_fee_or_rebate
            )

            price = (price_before_fees * self.state.base_spl_token_multiplier()) / (
                self.state.quote_spl_token_multiplier() * event.native_quantity_paid
            )

            size = event.native_quantity_paid / self.state.base_spl_token_multiplier()

        return t.FilledOrder(
            client_order_id=event.client_order_id,
            side=side,
            price=price,
            size=size,
            fee_cost=float(
                (Decimal(event.native_fee_or_rebate) / Decimal(self.state.base_spl_token_multiplier())) * Decimal(1 if event.event_flags.maker else -1)
            )
        )

    def _prepare_new_oo_account(
        self,
        owner: Keypair,
        balance_needed: int,
        signers: List[Keypair],
        transaction: Transaction,
    ) -> Pubkey:
        # new_open_orders_account = Account()
        new_open_orders_account = Keypair()
        place_order_open_order_account = new_open_orders_account.pubkey()
        transaction.add(
            make_create_account_instruction(
                owner_address=owner.pubkey(),
                new_account_address=new_open_orders_account.pubkey(),
                lamports=balance_needed,
                program_id=self.state.program_id(),
            )
        )
        signers.append(new_open_orders_account)
        return place_order_open_order_account

    def _prepare_order_transaction(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        transaction: Transaction,
        payer: Pubkey,
        owner: Keypair,
        order_type: OrderType,
        side: Side,
        signers: List[Keypair],
        limit_price: float,
        max_quantity: float,
        client_id: int,
        open_order_accounts: Union[
            List[OpenOrdersAccount], List[AsyncOpenOrdersAccount]
        ],
        place_order_open_order_account: Pubkey,
    ) -> None:
        # unwrapped SOL cannot be used for payment
        if payer == owner.pubkey():
            raise ValueError("Invalid payer account. Cannot use unwrapped SOL.")

        # TODO: add integration test for SOL wrapping.
        should_wrap_sol = (
            side == Side.BUY and self.state.quote_mint() == WRAPPED_SOL_MINT
        ) or (side == Side.SELL and self.state.base_mint() == WRAPPED_SOL_MINT)

        if should_wrap_sol:
            # wrapped_sol_account = Account()
            wrapped_sol_account = Keypair()
            payer = wrapped_sol_account.pubkey()
            signers.append(wrapped_sol_account)
            transaction.add(
                create_account(
                    CreateAccountParams(
                        from_pubkey=owner.pubkey(),
                        to_pubkey=wrapped_sol_account.pubkey(),
                        lamports=self._get_lamport_need_for_sol_wrapping(
                            limit_price, max_quantity, side, open_order_accounts
                        ),
                        space=ACCOUNT_LEN,
                        owner=TOKEN_PROGRAM_ID,
                    )
                )
            )
            transaction.add(
                initialize_account(
                    InitializeAccountParams(
                        account=wrapped_sol_account.pubkey(),
                        mint=WRAPPED_SOL_MINT,
                        owner=owner.pubkey(),
                        program_id=TOKEN_PROGRAM_ID,
                    )
                )
            )

        transaction.add(
            self.make_place_order_instruction(
                payer=payer,
                owner=owner,
                order_type=order_type,
                side=side,
                limit_price=limit_price,
                max_quantity=max_quantity,
                client_id=client_id,
                open_order_account=place_order_open_order_account,
            )
        )

        if should_wrap_sol:
            transaction.add(
                close_account(
                    CloseAccountParams(
                        account=wrapped_sol_account.pubkey(),
                        owner=owner.pubkey(),
                        dest=owner.pubkey(),
                        program_id=TOKEN_PROGRAM_ID,
                    )
                )
            )

    def _after_oo_mbfre_resp(
        self,
        mbfre_resp: GetMinimumBalanceForRentExemptionResp,
        owner: Keypair,
        signers: List[Keypair],
        transaction: Transaction,
    ) -> Pubkey:
        balance_needed = mbfre_resp.value
        place_order_open_order_account = self._prepare_new_oo_account(
            owner, balance_needed, signers, transaction
        )
        return place_order_open_order_account

    @staticmethod
    def _get_lamport_need_for_sol_wrapping(
        price: float,
        size: float,
        side: Side,
        open_orders_accounts: Union[
            List[OpenOrdersAccount], List[AsyncOpenOrdersAccount]
        ],
    ) -> int:
        lamports = 0
        if side == Side.BUY:
            lamports = round(price * size * 1.01 * LAMPORTS_PER_SOL)
            if open_orders_accounts:
                lamports -= open_orders_accounts[0].quote_token_free
        else:
            lamports = round(size * LAMPORTS_PER_SOL)
            if open_orders_accounts:
                lamports -= open_orders_accounts[0].base_token_free

        return max(lamports, 0) + 10000000

    def make_place_order_instruction(  # pylint: disable=too-many-arguments
        self,
        payer: Pubkey,
        owner: Keypair,
        order_type: OrderType,
        side: Side,
        limit_price: float,
        max_quantity: float,
        client_id: int,
        open_order_account: Pubkey,
        fee_discount_pubkey: Pubkey = None,
    ) -> Instruction:
        if self.state.base_size_number_to_lots(max_quantity) < 0:
            raise Exception(f"Size lot %d is too small {max_quantity}")
        if self.state.price_number_to_lots(limit_price) < 0:
            raise Exception(f"Price lot %d is too small {limit_price}")
        if self._use_request_queue():
            return instructions.new_order(
                instructions.NewOrderParams(
                    market=self.state.public_key(),
                    open_orders=open_order_account,
                    payer=payer,
                    owner=owner.pubkey(),
                    request_queue=self.state.request_queue(),
                    base_vault=self.state.base_vault(),
                    quote_vault=self.state.quote_vault(),
                    side=side,
                    limit_price=self.state.price_number_to_lots(limit_price),
                    max_quantity=self.state.base_size_number_to_lots(max_quantity),
                    order_type=order_type,
                    client_id=client_id,
                    program_id=self.state.program_id(),
                )
            )
        return instructions.new_order_v3(
            instructions.NewOrderV3Params(
                market=self.state.public_key(),
                open_orders=open_order_account,
                payer=payer,
                owner=owner.pubkey(),
                request_queue=self.state.request_queue(),
                event_queue=self.state.event_queue(),
                bids=self.state.bids(),
                asks=self.state.asks(),
                base_vault=self.state.base_vault(),
                quote_vault=self.state.quote_vault(),
                side=side,
                limit_price=self.state.price_number_to_lots(limit_price),
                max_base_quantity=self.state.base_size_number_to_lots(max_quantity),
                max_quote_quantity=self.state.base_size_number_to_lots(max_quantity)
                * self.state.quote_lot_size()
                * self.state.price_number_to_lots(limit_price),
                order_type=order_type,
                client_id=client_id,
                program_id=self.state.program_id(),
                self_trade_behavior=SelfTradeBehavior.DECREMENT_TAKE,
                fee_discount_pubkey=fee_discount_pubkey,
                limit=65535,
            )
        )

    def _build_cancel_order_by_client_id_tx(
        self, owner: Keypair, open_orders_account: Pubkey, client_id: int
    ) -> Transaction:
        return Transaction().add(
            self.make_cancel_order_by_client_id_instruction(
                owner, open_orders_account, client_id
            )
        )

    def make_cancel_order_by_client_id_instruction(
        self, owner: Keypair, open_orders_account: Pubkey, client_id: int
    ) -> Instruction:
        if self._use_request_queue():
            return instructions.cancel_order_by_client_id(
                instructions.CancelOrderByClientIDParams(
                    market=self.state.public_key(),
                    owner=owner.pubkey(),
                    open_orders=open_orders_account,
                    request_queue=self.state.request_queue(),
                    client_id=client_id,
                    program_id=self.state.program_id(),
                )
            )
        return instructions.cancel_order_by_client_id_v2(
            instructions.CancelOrderByClientIDV2Params(
                market=self.state.public_key(),
                owner=owner.pubkey(),
                open_orders=open_orders_account,
                bids=self.state.bids(),
                asks=self.state.asks(),
                event_queue=self.state.event_queue(),
                client_id=client_id,
                program_id=self.state.program_id(),
            )
        )

    def _build_cancel_order_tx(self, owner: Keypair, order: t.Order) -> Transaction:
        return Transaction().add(
            self.make_cancel_order_instruction(owner.pubkey(), order)
        )

    def make_cancel_order_instruction(
        self, owner: Pubkey, order: t.Order
    ) -> Instruction:
        if self._use_request_queue():
            return instructions.cancel_order(
                instructions.CancelOrderParams(
                    market=self.state.public_key(),
                    owner=owner,
                    open_orders=order.open_order_address,
                    request_queue=self.state.request_queue(),
                    side=order.side,
                    order_id=order.order_id,
                    open_orders_slot=order.open_order_slot,
                    program_id=self.state.program_id(),
                )
            )
        return instructions.cancel_order_v2(
            instructions.CancelOrderV2Params(
                market=self.state.public_key(),
                owner=owner,
                open_orders=order.open_order_address,
                bids=self.state.bids(),
                asks=self.state.asks(),
                event_queue=self.state.event_queue(),
                side=order.side,
                order_id=order.order_id,
                open_orders_slot=order.open_order_slot,
                program_id=self.state.program_id(),
            )
        )

    def _build_match_orders_tx(self, limit: int) -> Transaction:
        return Transaction().add(self.make_match_orders_instruction(limit))

    def make_match_orders_instruction(self, limit: int) -> Instruction:
        params = instructions.MatchOrdersParams(
            market=self.state.public_key(),
            request_queue=self.state.request_queue(),
            event_queue=self.state.event_queue(),
            bids=self.state.bids(),
            asks=self.state.asks(),
            base_vault=self.state.base_vault(),
            quote_vault=self.state.quote_vault(),
            limit=limit,
            program_id=self.state.program_id(),
        )
        return instructions.match_orders(params)

    def _build_settle_funds_tx(  # pylint: disable=too-many-arguments
        self,
        owner: Keypair,
        signers: List[Keypair],
        open_orders: Union[OpenOrdersAccount, AsyncOpenOrdersAccount],
        base_wallet: Pubkey,
        quote_wallet: Pubkey,  # TODO: add referrer_quote_wallet.
        min_bal_for_rent_exemption: int,
        should_wrap_sol: bool,
    ) -> Transaction:
        # TODO: Handle wrapped sol accounts
        if open_orders.owner != owner.pubkey():
            raise Exception("Invalid open orders account")
        vault_signer = Pubkey.create_program_address(
            [
                bytes(self.state.public_key()),
                self.state.vault_signer_nonce().to_bytes(8, byteorder="little"),
            ],
            self.state.program_id(),
        )
        transaction = Transaction()

        if should_wrap_sol:
            wrapped_sol_account = Keypair()
            signers.append(wrapped_sol_account)
            # make a wrapped SOL account with enough balance to
            # fund the trade, run the program, then send itself back home
            transaction.add(
                create_account(
                    CreateAccountParams(
                        from_pubkey=owner.pubkey(),
                        to_pubkey=wrapped_sol_account.pubkey(),
                        lamports=min_bal_for_rent_exemption,
                        space=ACCOUNT_LEN,
                        owner=TOKEN_PROGRAM_ID,
                    )
                )
            )
            # this was also broken upstream. it should be minting wrapped SOL, and using the token program ID
            transaction.add(
                initialize_account(
                    InitializeAccountParams(
                        account=wrapped_sol_account.pubkey(),
                        mint=WRAPPED_SOL_MINT,
                        owner=owner.pubkey(),
                        program_id=TOKEN_PROGRAM_ID,
                    )
                )
            )

        transaction.add(
            self.make_settle_funds_instruction(
                open_orders,
                base_wallet
                if self.state.base_mint() != WRAPPED_SOL_MINT
                else wrapped_sol_account.pubkey(),
                quote_wallet
                if self.state.quote_mint() != WRAPPED_SOL_MINT
                else wrapped_sol_account.pubkey(),
                vault_signer,
            )
        )

        if should_wrap_sol:
            # close out the account and send the funds home when the trade is completed/cancelled
            transaction.add(
                close_account(
                    CloseAccountParams(
                        account=wrapped_sol_account.pubkey(),
                        owner=owner.pubkey(),
                        dest=owner.pubkey(),
                        program_id=TOKEN_PROGRAM_ID,
                    )
                )
            )
        return transaction

    def _settle_funds_should_wrap_sol(self) -> bool:
        return (self.state.quote_mint() == WRAPPED_SOL_MINT) or (
            self.state.base_mint() == WRAPPED_SOL_MINT
        )

    def make_settle_funds_instruction(
        self,
        open_orders_account: Union[OpenOrdersAccount, AsyncOpenOrdersAccount],
        base_wallet: Pubkey,
        quote_wallet: Pubkey,
        vault_signer: Pubkey,
    ) -> Instruction:
        if base_wallet == self.state.base_vault():
            raise ValueError("base_wallet should not be a vault address")
        if quote_wallet == self.state.quote_vault():
            raise ValueError("quote_wallet should not be a vault address")

        return instructions.settle_funds(
            instructions.SettleFundsParams(
                market=self.state.public_key(),
                open_orders=open_orders_account.address,
                owner=open_orders_account.owner,
                base_vault=self.state.base_vault(),
                quote_vault=self.state.quote_vault(),
                base_wallet=base_wallet,
                quote_wallet=quote_wallet,
                vault_signer=vault_signer,
                program_id=self.state.program_id(),
            )
        )



import math
from enum import IntEnum
from typing import List, Optional, Tuple, Union, cast

from construct import Container
from solders.pubkey import Pubkey
from ..._layouts.queue import EVENT_LAYOUT, QUEUE_HEADER_LAYOUT, REQUEST_LAYOUT
from ..types import Event, EventFlags, Request, ReuqestFlags


class QueueType(IntEnum):
    EVENT = 1
    REQUEST = 2


def __from_bytes(
    buffer: bytes, queue_type: QueueType, history: Optional[int]
) -> Tuple[Container, List[Union[Event, Request]]]:
    header = QUEUE_HEADER_LAYOUT.parse(buffer)
    layout_size = (
        EVENT_LAYOUT.sizeof()
        if queue_type == QueueType.EVENT
        else REQUEST_LAYOUT.sizeof()
    )
    alloc_len = math.floor((len(buffer) - QUEUE_HEADER_LAYOUT.sizeof()) / layout_size)
    nodes: List[Union[Event, Request]] = []
    if history:
        for i in range(min(history, alloc_len)):
            node_index = (header.head + header.count + alloc_len - 1 - i) % alloc_len
            offset = QUEUE_HEADER_LAYOUT.sizeof() + node_index * layout_size
            nodes.append(
                __parse_queue_item(buffer[offset : offset + layout_size], queue_type)
            )  # noqa: E203
    else:
        for i in range(header.count):
            node_index = (header.head + i) % alloc_len
            offset = QUEUE_HEADER_LAYOUT.sizeof() + node_index * layout_size
            nodes.append(
                __parse_queue_item(buffer[offset : offset + layout_size], queue_type)
            )  # noqa: E203
    return header, nodes


def __parse_queue_item(buffer: bytes, queue_type: QueueType) -> Union[Event, Request]:
    if queue_type == QueueType.EVENT:  # pylint: disable=no-else-return
        parsed_item = EVENT_LAYOUT.parse(buffer)
        parsed_event_flags = parsed_item.event_flags
        event_flags = EventFlags(
            fill=parsed_event_flags.fill,
            out=parsed_event_flags.out,
            bid=parsed_event_flags.bid,
            maker=parsed_event_flags.maker,
        )

        return Event(
            event_flags=event_flags,
            open_order_slot=parsed_item.open_order_slot,
            fee_tier=parsed_item.fee_tier,
            native_quantity_released=parsed_item.native_quantity_released,
            native_quantity_paid=parsed_item.native_quantity_paid,
            native_fee_or_rebate=parsed_item.native_fee_or_rebate,
            order_id=int.from_bytes(parsed_item.order_id, "little"),
            public_key=Pubkey.from_bytes(parsed_item.public_key),
            client_order_id=parsed_item.client_order_id,
        )
    else:
        parsed_item = REQUEST_LAYOUT.parse(buffer)
        parsed_request_flags = parsed_item.request_flags
        request_flags = ReuqestFlags(
            new_order=parsed_request_flags.new_order,
            cancel_order=parsed_request_flags.cancel_order,
            bid=parsed_request_flags.bid,
            post_only=parsed_request_flags.post_only,
            ioc=parsed_request_flags.ioc,
        )

        return Request(
            request_flags=request_flags,
            open_order_slot=parsed_item.open_order_slot,
            fee_tier=parsed_item.fee_tier,
            max_base_size_or_cancel_id=parsed_item.max_base_size_or_cancel_id,
            native_quote_quantity_locked=parsed_item.native_quote_quantity_locked,
            order_id=int.from_bytes(parsed_item.order_id, "little"),
            open_orders=Pubkey.from_bytes(parsed_item.open_orders),
            client_order_id=parsed_item.client_order_id,
        )


def decode_request_queue(buffer: bytes, history: Optional[int] = None) -> List[Request]:
    header, nodes = __from_bytes(buffer, QueueType.REQUEST, history)
    if not header.account_flags.initialized or not header.account_flags.request_queue:
        raise Exception(
            "Invalid requests queue, either not initialized or not a request queue."
        )
    return cast(List[Request], nodes)


def decode_event_queue(buffer: bytes, history: Optional[int] = None) -> List[Event]:
    header, nodes = __from_bytes(buffer, QueueType.EVENT, history)
    if not header.account_flags.initialized or not header.account_flags.event_queue:
        raise Exception(
            "Invalid events queue, either not initialized or not a event queue."
        )
    return cast(List[Event], nodes)



from construct import (
    BitsInteger,
    BitsSwapped,
    BitStruct,
    Bytes,
    Const,
    Flag,
    Int8ul,
    Int32ul,
    Int64ul,
    Padding,
)
from construct import Struct as cStruct

from .account_flags import ACCOUNT_FLAGS_LAYOUT

QUEUE_HEADER_LAYOUT = cStruct(
    Padding(5),
    "account_flags" / ACCOUNT_FLAGS_LAYOUT,
    "head" / Int32ul,
    Padding(4),
    "count" / Int32ul,
    Padding(4),
    "next_seq_num" / Int32ul,
    Padding(4),
)

REQUEST_FLAGS_LAYOUT = BitsSwapped(  # Swap to little endian
    BitStruct(
        "new_order" / Flag,
        "cancel_order" / Flag,
        "bid" / Flag,
        "post_only" / Flag,
        "ioc" / Flag,
        Const(0, BitsInteger(3)),  # Padding
    )
)

REQUEST_LAYOUT = cStruct(
    "request_flags" / REQUEST_FLAGS_LAYOUT,
    "open_order_slot" / Int8ul,
    "fee_tier" / Int8ul,
    Padding(5),
    "max_base_size_or_cancel_id" / Int64ul,
    "native_quote_quantity_locked" / Int64ul,
    "order_id" / Bytes(16),
    "open_orders" / Bytes(32),
    "client_order_id" / Int64ul,
)

EVENT_FLAGS_LAYOUT = BitsSwapped(
    BitStruct(
        "fill" / Flag,
        "out" / Flag,
        "bid" / Flag,
        "maker" / Flag,
        Padding(4),
    )
)

EVENT_LAYOUT = cStruct(
    "event_flags" / EVENT_FLAGS_LAYOUT,
    "open_order_slot" / Int8ul,
    "fee_tier" / Int8ul,
    Padding(5),
    "native_quantity_released" / Int64ul,
    "native_quantity_paid" / Int64ul,
    "native_fee_or_rebate" / Int64ul,
    "order_id" / Bytes(16),
    "public_key" / Bytes(32),
    "client_order_id" / Int64ul,
)



from construct import BitsInteger, BitsSwapped, BitStruct, Const, Flag

# We will use a bitstruct with 64 bits instead of the widebits implementation in serum-js.
ACCOUNT_FLAGS_LAYOUT = BitsSwapped(  # Swap to little endian
    BitStruct(
        "initialized" / Flag,
        "market" / Flag,
        "open_orders" / Flag,
        "request_queue" / Flag,
        "event_queue" / Flag,
        "bids" / Flag,
        "asks" / Flag,
        Const(0, BitsInteger(57)),  # Padding
    )
)



from solders.pubkey import Pubkey
from solana.rpc.api import Client
from solders.account import Account
from solders.rpc.responses import GetAccountInfoResp
from spl.token.constants import WRAPPED_SOL_MINT

from pyserum._layouts.market import MINT_LAYOUT


def parse_bytes_data(res: GetAccountInfoResp) -> bytes:
    if not isinstance(res.value, Account):
        raise Exception("Cannot load byte data.")
    return res.value.data


def load_bytes_data(addr: Pubkey, conn: Client) -> bytes:
    res = conn.get_account_info(addr)
    return parse_bytes_data(res)


def parse_mint_decimals(bytes_data: bytes) -> int:
    return MINT_LAYOUT.parse(bytes_data).decimals


def get_mint_decimals(conn: Client, mint_pub_key: Pubkey) -> int:
    """Get the mint decimals for a token mint"""
    if mint_pub_key == WRAPPED_SOL_MINT:
        return 9

    bytes_data = load_bytes_data(mint_pub_key, conn)
    return parse_mint_decimals(bytes_data)



from typing import Any, Dict, List

import requests
from solders.pubkey import Pubkey
from solana.rpc.api import Client as conn  # pylint: disable=unused-import # noqa:F401

from .market.types import MarketInfo, TokenInfo

LIVE_MARKETS_URL = "https://raw.githubusercontent.com/project-serum/serum-ts/master/packages/serum/src/markets.json"
TOKEN_MINTS_URL = "https://raw.githubusercontent.com/project-serum/serum-ts/master/packages/serum/src/token-mints.json"


def parse_live_markets(data: List[Dict[str, Any]]) -> List[MarketInfo]:
    return [
        MarketInfo(name=m["name"], address=m["address"], program_id=m["programId"])
        for m in data
        if not m["deprecated"]
    ]


def parse_token_mints(data: List[Dict[str, str]]) -> List[TokenInfo]:
    return [
        TokenInfo(name=t["name"], address=Pubkey.from_string(t["address"]))
        for t in data
    ]


def get_live_markets() -> List[MarketInfo]:
    return parse_live_markets(requests.get(LIVE_MARKETS_URL).json())


def get_token_mints() -> List[TokenInfo]:
    return parse_token_mints(requests.get(TOKEN_MINTS_URL).json())



"""Layouts for dex instructions data."""
from enum import IntEnum

from construct import Bytes, Const, Int8ul, Int16ul, Int32ul, Int64ul, Pass
from construct import Struct as cStruct
from construct import Switch

from .slab import KEY


class InstructionType(IntEnum):
    INITIALIZE_MARKET = 0
    NEW_ORDER = 1
    MATCH_ORDER = 2
    CONSUME_EVENTS = 3
    CANCEL_ORDER = 4
    SETTLE_FUNDS = 5
    CANCEL_ORDER_BY_CLIENT_ID = 6
    NEW_ORDER_V3 = 10
    CANCEL_ORDER_V2 = 11
    CANCEL_ORDER_BY_CLIENT_ID_V2 = 12
    CLOSE_OPEN_ORDERS = 14
    INIT_OPEN_ORDERS = 15


_VERSION = 0

_INITIALIZE_MARKET = cStruct(
    "base_lot_size" / Int64ul,
    "quote_lot_size" / Int64ul,
    "fee_rate_bps" / Int16ul,
    "vault_signer_nonce" / Int64ul,
    "quote_dust_threshold" / Int64ul,
)

_NEW_ORDER = cStruct(
    "side" / Int32ul,  # Enum
    "limit_price" / Int64ul,
    "max_quantity" / Int64ul,
    "order_type" / Int32ul,  # Enum
    "client_id" / Int64ul,
)

_MATCH_ORDERS = cStruct("limit" / Int16ul)

_CONSUME_EVENTS = cStruct("limit" / Int16ul)

_CANCEL_ORDER = cStruct(
    "side" / Int32ul,  # Enum
    "order_id" / KEY,
    "open_orders" / Bytes(32),
    "open_orders_slot" / Int8ul,
)

_CANCEL_ORDER_BY_CLIENTID = cStruct("client_id" / Int64ul)

_NEW_ORDER_V3 = cStruct(
    "side" / Int32ul,  # Enum
    "limit_price" / Int64ul,
    "max_base_quantity" / Int64ul,
    "max_quote_quantity" / Int64ul,
    "self_trade_behavior" / Int32ul,
    "order_type" / Int32ul,  # Enum
    "client_id" / Int64ul,
    "limit" / Int16ul,
)

_CANCEL_ORDER_V2 = cStruct(
    "side" / Int32ul,  # Enum
    "order_id" / KEY,
)

_CANCEL_ORDER_BY_CLIENTID_V2 = cStruct("client_id" / Int64ul)

_CLOSE_OPEN_ORDERS = cStruct()
_INIT_OPEN_ORDERS = cStruct()

INSTRUCTIONS_LAYOUT = cStruct(
    "version" / Const(_VERSION, Int8ul),
    "instruction_type" / Int32ul,
    "args"
    / Switch(
        lambda this: this.instruction_type,
        {
            InstructionType.INITIALIZE_MARKET: _INITIALIZE_MARKET,
            InstructionType.NEW_ORDER: _NEW_ORDER,
            InstructionType.MATCH_ORDER: _MATCH_ORDERS,
            InstructionType.CONSUME_EVENTS: _CONSUME_EVENTS,
            InstructionType.CANCEL_ORDER: _CANCEL_ORDER,
            InstructionType.SETTLE_FUNDS: Pass,  # Empty list
            InstructionType.CANCEL_ORDER_BY_CLIENT_ID: _CANCEL_ORDER_BY_CLIENTID,
            InstructionType.NEW_ORDER_V3: _NEW_ORDER_V3,
            InstructionType.CANCEL_ORDER_V2: _CANCEL_ORDER_V2,
            InstructionType.CANCEL_ORDER_BY_CLIENT_ID_V2: _CANCEL_ORDER_BY_CLIENTID_V2,
            InstructionType.CLOSE_OPEN_ORDERS: _CLOSE_OPEN_ORDERS,
            InstructionType.INIT_OPEN_ORDERS: _INIT_OPEN_ORDERS,
        },
    ),
)



from __future__ import annotations

import math

from construct import Container, Struct
from solders.pubkey import Pubkey
from solana.rpc.api import Client
from solana.rpc.async_api import AsyncClient

from pyserum import async_utils, utils

from .._layouts.market import MARKET_LAYOUT
from .types import AccountFlags


class MarketState:  # pylint: disable=too-many-public-methods
    def __init__(
        self,
        parsed_market: Container,
        program_id: Pubkey,
        base_mint_decimals: int,
        quote_mint_decimals: int,
    ) -> None:
        self._decoded = parsed_market
        self._program_id = program_id
        self._base_mint_decimals = base_mint_decimals
        self._quote_mint_decimals = quote_mint_decimals

    @staticmethod
    def LAYOUT() -> Struct:  # pylint: disable=invalid-name
        """Construct layout of the market state."""
        return MARKET_LAYOUT

    @staticmethod
    def _make_parsed_market(bytes_data: bytes) -> Container:
        parsed_market = MARKET_LAYOUT.parse(bytes_data)
        # TODO: add ownAddress check!

        if (
            not parsed_market.account_flags.initialized
            or not parsed_market.account_flags.market
        ):
            raise Exception("Invalid market")
        return parsed_market

    @classmethod
    def load(
        cls, conn: Client, market_address: Pubkey, program_id: Pubkey
    ) -> MarketState:
        bytes_data = utils.load_bytes_data(market_address, conn)
        parsed_market = cls._make_parsed_market(bytes_data)

        base_mint_decimals = utils.get_mint_decimals(
            conn, Pubkey.from_bytes(parsed_market.base_mint)
        )
        quote_mint_decimals = utils.get_mint_decimals(
            conn, Pubkey.from_bytes(parsed_market.quote_mint)
        )
        return cls(parsed_market, program_id, base_mint_decimals, quote_mint_decimals)

    @classmethod
    async def async_load(
        cls, conn: AsyncClient, market_address: Pubkey, program_id: Pubkey
    ) -> MarketState:
        bytes_data = await async_utils.load_bytes_data(market_address, conn)
        parsed_market = cls._make_parsed_market(bytes_data)
        base_mint_decimals = await async_utils.get_mint_decimals(
            conn, Pubkey.from_bytes(parsed_market.base_mint)
        )
        quote_mint_decimals = await async_utils.get_mint_decimals(
            conn, Pubkey.from_bytes(parsed_market.quote_mint)
        )
        return cls(parsed_market, program_id, base_mint_decimals, quote_mint_decimals)

    @classmethod
    def from_bytes(
        cls,
        program_id: Pubkey,
        base_mint_decimals: int,
        quote_mint_decimals: int,
        buffer: bytes,
    ) -> MarketState:
        parsed_market = MARKET_LAYOUT.parse(buffer)
        # TODO: add ownAddress check!

        if (
            not parsed_market.account_flags.initialized
            or not parsed_market.account_flags.market
        ):
            raise Exception("Invalid market")

        return cls(parsed_market, program_id, base_mint_decimals, quote_mint_decimals)

    def program_id(self) -> Pubkey:
        return self._program_id

    def public_key(self) -> Pubkey:
        return Pubkey.from_bytes(self._decoded.own_address)

    def account_flags(self) -> AccountFlags:
        return AccountFlags(**self._decoded.account_flags)

    def asks(self) -> Pubkey:
        return Pubkey.from_bytes(self._decoded.asks)

    def bids(self) -> Pubkey:
        return Pubkey.from_bytes(self._decoded.bids)

    def fee_rate_bps(self) -> int:
        return self._decoded.fee_rate_bps

    def event_queue(self) -> Pubkey:
        return Pubkey.from_bytes(self._decoded.event_queue)

    def request_queue(self) -> Pubkey:
        return Pubkey.from_bytes(self._decoded.request_queue)

    def vault_signer_nonce(self) -> int:
        return self._decoded.vault_signer_nonce

    def base_mint(self) -> Pubkey:
        return Pubkey.from_bytes(self._decoded.base_mint)

    def quote_mint(self) -> Pubkey:
        return Pubkey.from_bytes(self._decoded.quote_mint)

    def base_vault(self) -> Pubkey:
        return Pubkey.from_bytes(self._decoded.base_vault)

    def quote_vault(self) -> Pubkey:
        return Pubkey.from_bytes(self._decoded.quote_vault)

    def base_deposits_total(self) -> int:
        return self._decoded.base_deposits_total

    def quote_deposits_total(self) -> int:
        return self._decoded.quote_deposits_total

    def base_fees_accrued(self) -> int:
        return self._decoded.base_fees_accrued

    def quote_fees_accrued(self) -> int:
        return self._decoded.quote_fees_accrued

    def quote_dust_threshold(self) -> int:
        return self._decoded.quote_dust_threshold

    def base_spl_token_decimals(self) -> int:
        return self._base_mint_decimals

    def quote_spl_token_decimals(self) -> int:
        return self._quote_mint_decimals

    def base_spl_token_multiplier(self) -> int:
        return 10**self._base_mint_decimals

    def quote_spl_token_multiplier(self) -> int:
        return 10**self._quote_mint_decimals

    def base_spl_size_to_number(self, size: int) -> float:
        return size / self.base_spl_token_multiplier()

    def quote_spl_size_to_number(self, size: int) -> float:
        return size / self.quote_spl_token_multiplier()

    def base_lot_size(self) -> int:
        return self._decoded.base_lot_size

    def quote_lot_size(self) -> int:
        return self._decoded.quote_lot_size

    def price_lots_to_number(self, price: int) -> float:
        return float(
            price * self.quote_lot_size() * self.base_spl_token_multiplier()
        ) / (self.base_lot_size() * self.quote_spl_token_multiplier())

    def price_number_to_lots(self, price: float) -> int:
        return int(
            round(
                (price * self.quote_spl_token_multiplier() * self.base_lot_size())
                / (self.base_spl_token_multiplier() * self.quote_lot_size())
            )
        )

    def base_size_lots_to_number(self, size: int) -> float:
        return float(size * self.base_lot_size()) / self.base_spl_token_multiplier()

    def base_size_number_to_lots(self, size: float) -> int:
        return int(
            math.floor(size * self.base_spl_token_multiplier()) / self.base_lot_size()
        )

    def quote_size_lots_to_number(self, size: int) -> float:
        return float(size * self.quote_lot_size()) / self.quote_spl_token_multiplier()

    def quote_size_number_to_lots(self, size: float) -> int:
        return int(
            math.floor(size * self.quote_spl_token_multiplier()) / self.quote_lot_size()
        )



from __future__ import annotations

from typing import List, NamedTuple, Tuple, Type, TypeVar

from solders.pubkey import Pubkey
from solana.rpc.api import Client
from solana.rpc.commitment import Processed
from solana.rpc.types import Commitment, MemcmpOpts
from solders.system_program import CreateAccountParams, create_account
from solders.instruction import Instruction
from solders.rpc.responses import GetProgramAccountsResp

from ._layouts.open_orders import OPEN_ORDERS_LAYOUT
from .instructions import DEFAULT_DEX_PROGRAM_ID
from .utils import load_bytes_data


class ProgramAccount(NamedTuple):
    public_key: Pubkey
    data: bytes
    is_executablable: bool
    lamports: int
    owner: Pubkey


_T = TypeVar("_T", bound="_OpenOrdersAccountCore")


class _OpenOrdersAccountCore:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        address: Pubkey,
        market: Pubkey,
        owner: Pubkey,
        base_token_free: int,
        base_token_total: int,
        quote_token_free: int,
        quote_token_total: int,
        free_slot_bits: int,
        is_bid_bits: int,
        orders: List[int],
        client_ids: List[int],
    ):
        self.address = address
        self.market = market
        self.owner = owner
        self.base_token_free = base_token_free
        self.base_token_total = base_token_total
        self.quote_token_free = quote_token_free
        self.quote_token_total = quote_token_total
        self.free_slot_bits = free_slot_bits
        self.is_bid_bits = is_bid_bits
        self.orders = orders
        self.client_ids = client_ids

    @classmethod
    def from_bytes(cls: Type[_T], address: Pubkey, buffer: bytes) -> _T:
        open_order_decoded = OPEN_ORDERS_LAYOUT.parse(buffer)
        if (
            not open_order_decoded.account_flags.open_orders
            or not open_order_decoded.account_flags.initialized
        ):
            raise Exception("Not an open order account or not initialized.")

        return cls(
            address=address,
            market=Pubkey.from_bytes(open_order_decoded.market),
            owner=Pubkey.from_bytes(open_order_decoded.owner),
            base_token_free=open_order_decoded.base_token_free,
            base_token_total=open_order_decoded.base_token_total,
            quote_token_free=open_order_decoded.quote_token_free,
            quote_token_total=open_order_decoded.quote_token_total,
            free_slot_bits=int.from_bytes(open_order_decoded.free_slot_bits, "little"),
            is_bid_bits=int.from_bytes(open_order_decoded.is_bid_bits, "little"),
            orders=[
                int.from_bytes(order, "little") for order in open_order_decoded.orders
            ],
            client_ids=open_order_decoded.client_ids,
        )

    @classmethod
    def _process_get_program_accounts_resp(
        cls: Type[_T], resp: GetProgramAccountsResp
    ) -> List[_T]:
        accounts = []
        for keyed_account in resp.value:
            account_details = keyed_account.account
            accounts.append(
                ProgramAccount(
                    public_key=keyed_account.pubkey,
                    data=account_details.data,
                    is_executablable=account_details.executable,
                    owner=account_details.owner,
                    lamports=account_details.lamports,
                )
            )

        return [
            cls.from_bytes(account.public_key, account.data) for account in accounts
        ]

    @staticmethod
    def _build_get_program_accounts_args(
        market: Pubkey,
        program_id: Pubkey,
        owner: Pubkey,
        commitment: Commitment,
    ) -> Tuple[Pubkey, Commitment, str, None, List[MemcmpOpts]]:
        filters = [
            MemcmpOpts(
                offset=5 + 8,  # 5 bytes of padding, 8 bytes of account flag
                bytes=str(market),
            ),
            MemcmpOpts(
                offset=5
                + 8
                + 32,  # 5 bytes of padding, 8 bytes of account flag, 32 bytes of market public key
                bytes=str(owner),
            ),
        ]
        data_slice = None
        return (
            program_id,
            commitment,
            "base64",
            data_slice,
            filters,
        )


class OpenOrdersAccount(_OpenOrdersAccountCore):
    @classmethod
    def find_for_market_and_owner(  # pylint: disable=too-many-arguments
        cls,
        conn: Client,
        market: Pubkey,
        owner: Pubkey,
        program_id: Pubkey,
        commitment: Commitment = Processed,
    ) -> List[OpenOrdersAccount]:
        args = cls._build_get_program_accounts_args(
            market=market, program_id=program_id, owner=owner, commitment=commitment
        )
        resp = conn.get_program_accounts(*args)
        return cls._process_get_program_accounts_resp(resp)

    @classmethod
    def load(cls, conn: Client, address: str) -> OpenOrdersAccount:
        addr_pub_key = Pubkey.from_string(address)
        bytes_data = load_bytes_data(addr_pub_key, conn)
        return cls.from_bytes(addr_pub_key, bytes_data)


def make_create_account_instruction(
    owner_address: Pubkey,
    new_account_address: Pubkey,
    lamports: int,
    program_id: Pubkey = DEFAULT_DEX_PROGRAM_ID,
) -> Instruction:
    return create_account(
        CreateAccountParams(
            from_pubkey=owner_address,
            to_pubkey=new_account_address,
            lamports=lamports,
            space=OPEN_ORDERS_LAYOUT.sizeof(),
            owner=program_id,
        )
    )



"""Market module to interact with Serum DEX."""
from __future__ import annotations

from typing import List

from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.types import TxOpts
from solana.transaction import Transaction
from solders.rpc.responses import SendTransactionResp

import pyserum.market.types as t
from pyserum import instructions

from .._layouts.open_orders import OPEN_ORDERS_LAYOUT
from ..async_open_orders_account import AsyncOpenOrdersAccount
from ..async_utils import load_bytes_data
from ..enums import OrderType, Side
from ._internal.queue import decode_event_queue, decode_request_queue
from .core import MarketCore
from .orderbook import OrderBook
from .state import MarketState

LAMPORTS_PER_SOL = 1000000000


# pylint: disable=too-many-public-methods,abstract-method
class AsyncMarket(MarketCore):
    """Represents a Serum Market."""

    def __init__(
        self,
        conn: AsyncClient,
        market_state: MarketState,
        force_use_request_queue: bool = False,
    ) -> None:
        super().__init__(
            market_state=market_state, force_use_request_queue=force_use_request_queue
        )
        self._conn = conn

    @classmethod
    # pylint: disable=unused-argument
    async def load(
        cls,
        conn: AsyncClient,
        market_address: Pubkey,
        program_id: Pubkey = instructions.DEFAULT_DEX_PROGRAM_ID,
        force_use_request_queue: bool = False,
    ) -> AsyncMarket:
        """Factory method to create a Market.

        :param conn: The connection that we use to load the data, created from `solana.rpc.api`.
        :param market_address: The market address that you want to connect to.
        :param program_id: The program id of the given market, it will use the default value if not provided.
        """
        market_state = await MarketState.async_load(conn, market_address, program_id)
        return cls(conn, market_state, force_use_request_queue)

    async def find_open_orders_accounts_for_owner(
        self, owner_address: Pubkey
    ) -> List[AsyncOpenOrdersAccount]:
        return await AsyncOpenOrdersAccount.find_for_market_and_owner(
            self._conn, self.state.public_key(), owner_address, self.state.program_id()
        )

    async def load_bids(self) -> OrderBook:
        """Load the bid order book"""
        bytes_data = await load_bytes_data(self.state.bids(), self._conn)
        return self._parse_bids_or_asks(bytes_data)

    async def load_asks(self) -> OrderBook:
        """Load the ask order book."""
        bytes_data = await load_bytes_data(self.state.asks(), self._conn)
        return self._parse_bids_or_asks(bytes_data)

    async def load_orders_for_owner(self, owner_address: Pubkey) -> List[t.Order]:
        """Load orders for owner."""
        bids = await self.load_bids()
        asks = await self.load_asks()
        open_orders_accounts = await self.find_open_orders_accounts_for_owner(
            owner_address
        )
        return self._parse_orders_for_owner(bids, asks, open_orders_accounts)

    async def load_event_queue(self) -> List[t.Event]:
        """Load the event queue which includes the fill item and out item. For any trades two fill items are added to
        the event queue. And in case of a trade, cancel or IOC order that missed, out items are added to the event
        queue.
        """
        bytes_data = await load_bytes_data(self.state.event_queue(), self._conn)
        return decode_event_queue(bytes_data)

    async def load_request_queue(self) -> List[t.Request]:
        bytes_data = await load_bytes_data(self.state.request_queue(), self._conn)
        return decode_request_queue(bytes_data)

    async def load_fills(self, limit=100) -> List[t.FilledOrder]:
        bytes_data = await load_bytes_data(self.state.event_queue(), self._conn)
        return self._parse_fills(bytes_data, limit)

    async def place_order(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        payer: Pubkey,
        owner: Keypair,
        order_type: OrderType,
        side: Side,
        limit_price: float,
        max_quantity: float,
        client_id: int = 0,
        opts: TxOpts = TxOpts(),
    ) -> SendTransactionResp:  # TODO: Add open_orders_address_key param and fee_discount_pubkey
        transaction = Transaction()
        signers: List[Keypair] = [owner]
        open_order_accounts = await self.find_open_orders_accounts_for_owner(
            owner.pubkey()
        )
        if open_order_accounts:
            place_order_open_order_account = open_order_accounts[0].address
        else:
            mbfre_resp = await self._conn.get_minimum_balance_for_rent_exemption(
                OPEN_ORDERS_LAYOUT.sizeof()
            )
            place_order_open_order_account = self._after_oo_mbfre_resp(
                mbfre_resp=mbfre_resp,
                owner=owner,
                signers=signers,
                transaction=transaction,
            )
            # TODO: Cache new_open_orders_account
        # TODO: Handle fee_discount_pubkey

        self._prepare_order_transaction(
            transaction=transaction,
            payer=payer,
            owner=owner,
            order_type=order_type,
            side=side,
            signers=signers,
            limit_price=limit_price,
            max_quantity=max_quantity,
            client_id=client_id,
            open_order_accounts=open_order_accounts,
            place_order_open_order_account=place_order_open_order_account,
        )
        return await self._conn.send_transaction(transaction, *signers, opts=opts)

    async def cancel_order_by_client_id(
        self,
        owner: Keypair,
        open_orders_account: Pubkey,
        client_id: int,
        opts: TxOpts = TxOpts(),
    ) -> SendTransactionResp:
        txs = self._build_cancel_order_by_client_id_tx(
            owner=owner, open_orders_account=open_orders_account, client_id=client_id
        )
        return await self._conn.send_transaction(txs, owner, opts=opts)

    async def cancel_order(
        self, owner: Keypair, order: t.Order, opts: TxOpts = TxOpts()
    ) -> SendTransactionResp:
        txn = self._build_cancel_order_tx(owner=owner, order=order)
        return await self._conn.send_transaction(txn, owner, opts=opts)

    async def match_orders(
        self, fee_payer: Keypair, limit: int, opts: TxOpts = TxOpts()
    ) -> SendTransactionResp:
        txn = self._build_match_orders_tx(limit)
        return await self._conn.send_transaction(txn, fee_payer, opts=opts)

    async def settle_funds(  # pylint: disable=too-many-arguments
        self,
        owner: Keypair,
        open_orders: AsyncOpenOrdersAccount,
        base_wallet: Pubkey,
        quote_wallet: Pubkey,  # TODO: add referrer_quote_wallet.
        opts: TxOpts = TxOpts(),
    ) -> SendTransactionResp:
        # TODO: Handle wrapped sol accounts
        should_wrap_sol = self._settle_funds_should_wrap_sol()
        if should_wrap_sol:
            mbfre_resp = await self._conn.get_minimum_balance_for_rent_exemption(165)
            min_bal_for_rent_exemption = mbfre_resp.value
        else:
            min_bal_for_rent_exemption = 0  # value only matters if should_wrap_sol
        signers = [owner]
        transaction = self._build_settle_funds_tx(
            owner=owner,
            signers=signers,
            open_orders=open_orders,
            base_wallet=base_wallet,
            quote_wallet=quote_wallet,
            min_bal_for_rent_exemption=min_bal_for_rent_exemption,
            should_wrap_sol=should_wrap_sol,
        )
        return await self._conn.send_transaction(transaction, *signers, opts=opts)



"""Market module to interact with Serum DEX."""
from __future__ import annotations

from typing import List

from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solana.rpc.api import Client
from solana.rpc.types import TxOpts
from solana.transaction import Transaction
from solders.rpc.responses import SendTransactionResp

import pyserum.market.types as t
from pyserum import instructions

from .._layouts.open_orders import OPEN_ORDERS_LAYOUT
from ..enums import OrderType, Side
from ..open_orders_account import OpenOrdersAccount
from ..utils import load_bytes_data
from ._internal.queue import decode_event_queue, decode_request_queue
from .core import MarketCore
from .orderbook import OrderBook
from .state import MarketState

LAMPORTS_PER_SOL = 1000000000


# pylint: disable=too-many-public-methods,abstract-method
class Market(MarketCore):
    """Represents a Serum Market."""

    def __init__(
        self,
        conn: Client,
        market_state: MarketState,
        force_use_request_queue: bool = False,
    ) -> None:
        super().__init__(
            market_state=market_state, force_use_request_queue=force_use_request_queue
        )
        self._conn = conn

    @classmethod
    # pylint: disable=unused-argument
    def load(
        cls,
        conn: Client,
        market_address: Pubkey,
        program_id: Pubkey = instructions.DEFAULT_DEX_PROGRAM_ID,
        force_use_request_queue: bool = False,
    ) -> Market:
        """Factory method to create a Market.

        :param conn: The connection that we use to load the data, created from `solana.rpc.api`.
        :param market_address: The market address that you want to connect to.
        :param program_id: The program id of the given market, it will use the default value if not provided.
        """
        market_state = MarketState.load(conn, market_address, program_id)
        return cls(conn, market_state, force_use_request_queue)

    def find_open_orders_accounts_for_owner(
        self, owner_address: Pubkey
    ) -> List[OpenOrdersAccount]:
        return OpenOrdersAccount.find_for_market_and_owner(
            self._conn, self.state.public_key(), owner_address, self.state.program_id()
        )

    def load_bids(self) -> OrderBook:
        """Load the bid order book"""
        bytes_data = load_bytes_data(self.state.bids(), self._conn)
        return self._parse_bids_or_asks(bytes_data)

    def load_asks(self) -> OrderBook:
        """Load the ask order book."""
        bytes_data = load_bytes_data(self.state.asks(), self._conn)
        return self._parse_bids_or_asks(bytes_data)

    def load_orders_for_owner(self, owner_address: Pubkey) -> List[t.Order]:
        """Load orders for owner."""
        bids = self.load_bids()
        asks = self.load_asks()
        open_orders_accounts = self.find_open_orders_accounts_for_owner(owner_address)
        return self._parse_orders_for_owner(bids, asks, open_orders_accounts)

    def load_event_queue(self) -> List[t.Event]:
        """Load the event queue which includes the fill item and out item. For any trades two fill items are added to
        the event queue. And in case of a trade, cancel or IOC order that missed, out items are added to the event
        queue.
        """
        bytes_data = load_bytes_data(self.state.event_queue(), self._conn)
        return decode_event_queue(bytes_data)

    def load_request_queue(self) -> List[t.Request]:
        bytes_data = load_bytes_data(self.state.request_queue(), self._conn)
        return decode_request_queue(bytes_data)

    def load_fills(self, limit=100) -> List[t.FilledOrder]:
        bytes_data = load_bytes_data(self.state.event_queue(), self._conn)
        return self._parse_fills(bytes_data, limit)

    def place_order(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        payer: Pubkey,
        owner: Keypair,
        order_type: OrderType,
        side: Side,
        limit_price: float,
        max_quantity: float,
        client_id: int = 0,
        opts: TxOpts = TxOpts(),
    ) -> SendTransactionResp:  # TODO: Add open_orders_address_key param and fee_discount_pubkey
        transaction = Transaction()
        signers: List[Keypair] = [owner]
        open_order_accounts = self.find_open_orders_accounts_for_owner(owner.pubkey())
        if open_order_accounts:
            place_order_open_order_account = open_order_accounts[0].address
        else:
            mbfre_resp = self._conn.get_minimum_balance_for_rent_exemption(
                OPEN_ORDERS_LAYOUT.sizeof()
            )
            place_order_open_order_account = self._after_oo_mbfre_resp(
                mbfre_resp=mbfre_resp,
                owner=owner,
                signers=signers,
                transaction=transaction,
            )
            # TODO: Cache new_open_orders_account
        # TODO: Handle fee_discount_pubkey

        self._prepare_order_transaction(
            transaction=transaction,
            payer=payer,
            owner=owner,
            order_type=order_type,
            side=side,
            signers=signers,
            limit_price=limit_price,
            max_quantity=max_quantity,
            client_id=client_id,
            open_order_accounts=open_order_accounts,
            place_order_open_order_account=place_order_open_order_account,
        )
        return self._conn.send_transaction(transaction, *signers, opts=opts)

    def cancel_order_by_client_id(
        self,
        owner: Keypair,
        open_orders_account: Pubkey,
        client_id: int,
        opts: TxOpts = TxOpts(),
    ) -> SendTransactionResp:
        txs = self._build_cancel_order_by_client_id_tx(
            owner=owner, open_orders_account=open_orders_account, client_id=client_id
        )
        return self._conn.send_transaction(txs, owner, opts=opts)

    def cancel_order(
        self, owner: Keypair, order: t.Order, opts: TxOpts = TxOpts()
    ) -> SendTransactionResp:
        txn = self._build_cancel_order_tx(owner=owner, order=order)
        return self._conn.send_transaction(txn, owner, opts=opts)

    def match_orders(
        self, fee_payer: Keypair, limit: int, opts: TxOpts = TxOpts()
    ) -> SendTransactionResp:
        txn = self._build_match_orders_tx(limit)
        return self._conn.send_transaction(txn, fee_payer, opts=opts)

    def settle_funds(  # pylint: disable=too-many-arguments
        self,
        owner: Keypair,
        open_orders: OpenOrdersAccount,
        base_wallet: Pubkey,
        quote_wallet: Pubkey,  # TODO: add referrer_quote_wallet.
        opts: TxOpts = TxOpts(),
    ) -> SendTransactionResp:
        # TODO: Handle wrapped sol accounts
        should_wrap_sol = self._settle_funds_should_wrap_sol()
        min_bal_for_rent_exemption = (
            self._conn.get_minimum_balance_for_rent_exemption(165).value
            if should_wrap_sol
            else 0
        )  # value only matters if should_wrap_sol
        signers = [owner]
        transaction = self._build_settle_funds_tx(
            owner=owner,
            signers=signers,
            open_orders=open_orders,
            base_wallet=base_wallet,
            quote_wallet=quote_wallet,
            min_bal_for_rent_exemption=min_bal_for_rent_exemption,
            should_wrap_sol=should_wrap_sol,
        )
        return self._conn.send_transaction(transaction, *signers, opts=opts)
